/******************************************************************************
Software License Agreement

 
    (c) 2016 Microchip Technology Inc. and its subsidiaries. 

    本軟體是由 Microchip 公司所提供的範例程式。 Microchip 不保證其功能性的完整，
    且本範例程式僅是用來展示其功能，對於任何侵權或相關版權及軟體的錯誤所導致損失
    ，損害，費用和開支等 Microchip 均不具賠償責任。

    注意 : 您必須接收上述條款後方能使用此範例程式。
 
*******************************************************************************/

#include <xc.h>
#include "lcd.h"                    // LCD 函數原型宣告
#include "mcc_generated_files/mcc.h"
 

// 底下的延遲時間的定義是在 Fsys = 20MHz 時所算出的數值
//
// Define the LCD Long Delay time in terms of loop time
// typically 1.64mS , Execute both Clear Display & Cursor Home functions
//
#define LCD_L_Delay         6675 // 2.5mS

// Define the LCD Short Delay time in terms of loop time
// typically 40uS, Here is define 100uS for 267 loop time
#define LCD_S_delay         267  // 0.1mS

// Define the startup time for the LCD in terms of loop time
// typically 10ms 
#define LCD_STARTUP         26700 // 10mS

#define LCD_MAX_COLUMN      16

#define LCD_COMMAND_CLEAR_SCREEN        0x01
#define LCD_COMMAND_RETURN_HOME         0x02
#define LCD_COMMAND_ENTER_DATA_MODE     0x06
#define LCD_COMMAND_CURSOR_OFF          0x0C
#define LCD_COMMAND_CURSOR_ON           0x0F
#define LCD_COMMAND_MOVE_CURSOR_LEFT    0x10
#define LCD_COMMAND_MOVE_CURSOR_RIGHT   0x14
#define LCD_COMMAND_SET_MODE_4_BIT      0x28
#define LCD_COMMAND_SET_MODE_8_BIT      0x38
#define LCD_COMMAND_ROW_0_HOME          0x80
#define LCD_COMMAND_ROW_1_HOME          0xC0
#define LCD_START_UP_COMMAND_1          0x33    
#define LCD_START_UP_COMMAND_2          0x32    

#define LCD_DATA_LAT                   LATB     // RB10:RB13 做為 LED4 ~ LED7 的資料 Bus
#define LCD_DATA_TRIS                  TRISB

// 底下有關 LCD 控制腳位已在 pin_manager.h 中宣告
// 其腳位名稱的設定是由 MCC 的 Pin Module 所輸入的腳位名稱
//
//#define LCD_RS_SetHigh()              LATBbits.LATB3 = 1 //set Register Select bit
//#define LCD_RS_SetLow()               LATBbits.LATB3 = 0 //clear Register Select bit
//#define LCD_RW_SetHigh()              LATBbits.LATB6 = 1  //set Read/Write bit
//#define LCD_RW_SetLow()               LATBbits.LATB6 = 0  //clear Read/Write bit
//#define LCD_E_SetHigh()               LATBbits.LATB5 = 1  //set Enable bit
//#define LCD_E_SetLow()                LATBbits.LATB5 = 0  //clear Enable bit

//#define LCD_RS_SetDigitalInpu()       TRISBbits.TRISB3 = 1 //set Register Select bit
//#define LCD_RS_SetDigitalOutput()     TRISBbits.TRISB3 = 0 //clear Register Select bit
//#define LCD_RW_SetDigitalInput()      TRISBbits.TRISB6 = 1  //set Read/Write bit
//#define LCD_RW_SetDigitalOutput()     TRISBbits.TRISB6 = 0  //clear Read/Write bit
//#define LCD_E_SetDigitalInput()       TRISBbits.TRISB5 = 1  //set Enable bit
//#define LCD_E_SetDigitalOutput()      TRISBbits.TRISB5 = 0  //clear Enable bit

/* Private variables ************************************************/
static uint8_t row ;
static uint8_t column ;
/*********************************************************************
 * Function: void LCD_Initialize(void);
 *
 * Overview: Initializes the LCD screen.  Can take several hundred
 *           milliseconds.
 *
 * PreCondition: none
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
void LCD_Initialize ( void )
{   
    // LCD 資料腳設定
    LCD_DATA_LAT &= 0xC3FF ;                // 設定 RB10 ~ RB13 為 LCD D4 ~ D7 輸出腳位
    LCD_DATA_TRIS &= 0xC3FF;
    
    // LCD 控制腳設定
    LCD_RW_SetLow ( ) ;                     // LCD R/W 腳為 0，輸出功能    
    LCD_RW_SetDigitalOutput ( )  ;
    LCD_RS_SetLow ( ) ;                     // LCD RS signal
    LCD_RS_SetDigitalOutput ( )  ;
    LCD_E_SetLow ( ) ;                      // LCD E signal
    LCD_E_SetDigitalOutput ( ) ;

    LCD_Wait ( LCD_STARTUP ) ;              // 電源開機，等待 10mS 的穩定時間
    
    LCD_SendCommand ( LCD_START_UP_COMMAND_1 ) ;    // 送出 0x3, 0x3 ,0x3 ,0x2 
    LCD_SendCommand ( LCD_START_UP_COMMAND_2 ) ;    // 完成 4-bit Mode 的設定 
      
    LCD_SendCommand ( LCD_COMMAND_SET_MODE_4_BIT  ) ;   // LCD 為 2 Line & 4 Bit
    LCD_SendCommand ( LCD_COMMAND_CURSOR_OFF  ) ;
    LCD_SendCommand ( LCD_COMMAND_ENTER_DATA_MODE ) ;

    LCD_ClearScreen ( ) ;                           // 清除 LCD 的顯示，游標歸零 

}

/*********************************************************************
 * 函數 : void LCD_Set_Cursor(unsigned char CurY, unsigned char CurX);
 *
 * 說明 : 設定 LCD 游標的位置
 *   
 * 用法 : CurY 為行別，範圍為 0 ~ 1，CurX 為欄位，範圍為 0 ~ 15
 *        LCD_Set_Cursor(1,6) 功能為設定游標位置在
 *        (第二行，第七個欄位) 做顯示
 * 
 * Output: None
 *
 ********************************************************************/
void LCD_Set_Cursor(unsigned char CurY, unsigned char CurX)
{
	LCD_SendCommand( 0x80 + CurY * 0x40 + CurX) ;
}

/*********************************************************************
 * 函數 : void LCD_PutString(char* inputString, uint16_t length);
 *
 * 說明 : 將放在 RAM 的字串顯示在 LCD 上.  顯示方式可以依所傳入的字串的長度做顯示
 *        或檢視出字串的結尾資料 0x00 (NULL) 後停止顯示。
 *
 * Input: char* - 要顯示的字串
 *        uint16_t - 字串的長度
 *
 * Output: None
 *
 ********************************************************************/
void LCD_PutString ( char* inputString , uint16_t length )
{
    while (length--)
    {
        switch (*inputString)
        {
            case 0x00:
                return ;

            default:
                LCD_PutChar ( *inputString++ ) ;
                break ;
        }
    }
}

/*********************************************************************
 * 函數 : void putsLCD( char *Str );
 *
 * 說明 : 顯示 RAM 記憶體所儲存的字串資料 
 *   
 * 用法 : 傳入必須是指標變數
 * 
 * Output: None
 *
 ********************************************************************/
void putsLCD( char *Str)
{
   char Str_Temp;
   
   while (1)
   {
	Str_Temp = *Str ;

		if (Str_Temp != 0x00 )
		   {
			LCD_SendData(Str_Temp) ;
			Str ++ ;
		   }
		else
			return ;
   }
}

/*********************************************************************
 * 函數 : void putrsLCD( const char *Str );
 *
 * 說明 : 顯示 Flash 記憶體所儲存的字串資料 (查表資料)
 *   
 * 用法 : 傳入的指標變數必須為 const 的宣告
 * 
 * Output: None
 *
 ********************************************************************/
void putrsLCD( const char *Str )
{
   char Str_Temp;
   
   while (1)
   {
	Str_Temp = *Str ;

		if (Str_Temp != 0x00 )
		   {
			LCD_SendData(Str_Temp) ;
			Str ++ ;
		   }
		else
			return ;
   }
}
/*********************************************************************
 * Function: void LCD_PutChar(char);
 *
 * Overview: Puts a character on the LCD screen.  Unsupported characters will be
 *           discarded.  May block or throw away characters is LCD is not ready
 *           or buffer space is not available.
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: char - character to print
 *
 * Output: None
 *
 ********************************************************************/
void LCD_PutChar ( char inputCharacter )
{
    switch (inputCharacter)
    {
        case '\r':
            LCD_CarriageReturn ( ) ;
            break ;

        case '\n':
            if (row == 0)
            {
                LCD_ShiftCursorDown ( ) ;
            }
            else
            {
                LCD_ShiftCursorUp ( ) ;
            }
            break ;

        case '\b':
            LCD_ShiftCursorLeft ( ) ;
            LCD_PutChar ( ' ' ) ;
            LCD_ShiftCursorLeft ( ) ;
            break ;

        default:
            LCD_SendData ( inputCharacter ) ;
            column++ ;

            if (column == LCD_MAX_COLUMN)
            {
                column = 0 ;
                if (row == 0)
                {
                    LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME ) ;
                    row = 1 ;
                }
                else
                {
                    LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME ) ;
                    row = 0 ;
                }
            }
            break ;
    }
}

/*********************************************************************
 * 函數 : void putcLCD(char LCD_Char);
 *
 * 說明 : 將一個字元或 8-bit 數送到 LCD 顯示。輸入字元需為 ASCII Code
 *        可顯示字元。
 *  
 * 用法 : putcLCD ('A'); putcLCD (0x41); putcLCD ('\n') 等輸入方式
 * 
 * Output: None
 *
 ********************************************************************/
void putcLCD(char LCD_Char)
{
	LCD_SendData(LCD_Char) ;
}

/*********************************************************************
 * 函數 : void puthexLCD(unsigned char HEX_Val)
 *
 * 說明 : 將一個字元或 8-bit 數送到 LCD 做 HEX 數值的顯示。
 *  
 * 用法 : puthexLCD (0x3a) 後會在 LCD 上的游標位制顯示 3A
 * 
 * Output: None
 *
 ********************************************************************/
void puthexLCD(unsigned char HEX_Val)
{
	unsigned char Temp_HEX ;

	Temp_HEX = (HEX_Val >> 4) & 0x0f ;

	if ( Temp_HEX > 9 )Temp_HEX += 0x37 ;
    else Temp_HEX += 0x30 ;

	LCD_SendData (Temp_HEX) ;

	Temp_HEX = HEX_Val  & 0x0f ;
	if ( Temp_HEX > 9 )Temp_HEX += 0x37 ;
    else Temp_HEX += 0x30 ;

	LCD_SendData (Temp_HEX) ;
}

/*********************************************************************
 * Function: void LCD_ClearScreen(void);
 *
 * Overview: Clears the screen, if possible.
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
void LCD_ClearScreen ( void )
{
    LCD_SendCommand ( LCD_COMMAND_CLEAR_SCREEN ) ;
    LCD_SendCommand ( LCD_COMMAND_RETURN_HOME ) ;

    row = 0 ;
    column = 0 ;
}

/*********************************************************************
 * Function: static void LCD_SendData(char data)
 *
 * Overview: Sends data to LCD
 *
 * PreCondition: None
 *
 * Input: char - contains the data to be sent to the LCD
 *
 * Output: None
 *
 ********************************************************************/
 void LCD_SendData ( char data )
{
    //Split data into nibbles 
    int data_shift = data << 8;
    int upper_nibble = data_shift & 0xF000;
    upper_nibble = upper_nibble >>2;
    int lower_nibble = data_shift & 0x0F00;
    lower_nibble = lower_nibble <<2;
    
    LCD_RW_SetLow ( ) ;                     // LCD R/W & LCD RS = Low
    LCD_RS_SetHigh ( ) ;                    // LCD 資料寫入設定
    LCD_DATA_LAT &= 0xC3FF ;
    LCD_DATA_LAT |= upper_nibble ;
    LCD_Enable_Timing( );
    
    LCD_DATA_LAT &= 0xC3FF ;
    LCD_DATA_LAT |= lower_nibble ;
    LCD_Enable_Timing( );
}
/*********************************************************************
 * Function: static void LCD_SendCommand(char data)
 *
 * Overview: Sends command to LCD
 *
 * PreCondition: None
 *
 * Input: char - contains the command to be sent to the LCD
 *        unsigned int - has the specific delay for the command
 *
 * Output: None
 *
 ********************************************************************/
void LCD_SendCommand ( char command )
{
    // 因為 LCD 的輸出腳位為 RB10 ~ RB13，傳進來的 8-bit 資料要做位移後並轉成
    // 4-bit 的 High-Nibble & Low-Nibble 後依序傳送到 LCD Module
    //
    int cmd_shift = command << 8;           // 傳入值為 0x31 ==> 0x3100
    int upper_nibble = cmd_shift & 0xF000;  // 0x3000
    int lower_nibble = cmd_shift & 0x0F00;  // 0x0100
    upper_nibble = upper_nibble >>2;        // 將 0x3 轉到 b10 ~ b13 位元後暫存
    lower_nibble = lower_nibble <<2;        // 將 0x1 轉到 b10 ~ b13 位元後暫存
    
    LCD_RW_SetLow ( ) ;                     // LCD R/W & LCD RS = Low
    LCD_RS_SetLow ( ) ;                     // LCD 命令寫入設定
    LCD_DATA_LAT &= 0xC3FF ;                // 將 4-bit High Nibble 送到 RB10 ~ RB13
    LCD_DATA_LAT |= upper_nibble ;
    LCD_Enable_Timing( );
    LCD_Wait(LCD_L_Delay);                  // Command Write delay 2.5mS
    
    LCD_DATA_LAT &= 0xC3FF ;                // 將 4-bit Low Nibble 送到 RB10 ~ RB13
    LCD_DATA_LAT |= lower_nibble ;
    LCD_Enable_Timing( );
    LCD_Wait(LCD_L_Delay);                  // Command Write delay 2.5mS
}

void LCD_Enable_Timing(void)
{
    Nop();                                  // Delay for data stable
    Nop();
    LCD_E_SetHigh ( ) ;                     // LCD Enable 啟動
    Nop();
    Nop();
    LCD_E_SetLow ( ) ;                      // LCD Disable
    LCD_Wait(LCD_S_delay);
}

/*********************************************************************
 * Function: static void LCD_Wait(unsigned int B)
 *
 * Overview: A crude wait function that just cycle burns
 *
 * PreCondition: None
 *
 * Input: unsigned int - artibrary delay time based on loop counts.
 *
 * Output: None
 *
 ********************************************************************/
void LCD_Wait ( unsigned int delay )
{
    while (delay)
    {
        delay-- ;
    }
}

/*********************************************************************
 * Function: void LCD_CursorEnable(bool enable)
 *
 * Overview: Enables/disables the cursor
 *
 * PreCondition: None
 *
 * Input: bool - specifies if the cursor should be on or off
 *
 * Output: None
 *
 ********************************************************************/
void LCD_CursorEnable ( bool enable )
{
    if (enable == true)
    {
        LCD_SendCommand ( LCD_COMMAND_CURSOR_ON ) ;
    }
    else
    {
        LCD_SendCommand ( LCD_COMMAND_CURSOR_OFF) ;
    }
}

/*******************************************************************/
/*******************************************************************/
/* Private Functions ***********************************************/
/*******************************************************************/
/*******************************************************************/
/*********************************************************************
 * Function: static void LCD_CarriageReturn(void)
 *
 * Overview: Handles a carriage return
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
static void LCD_CarriageReturn ( void )
{
    if (row == 0)
    {
        LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME ) ;
    }
    else
    {
        LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME  ) ;
    }
    column = 0 ;
}
/*********************************************************************
 * Function: static void LCD_ShiftCursorLeft(void)
 *
 * Overview: Shifts cursor left one spot (wrapping if required)
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
static void LCD_ShiftCursorLeft ( void )
{
    uint8_t i ;

    if (column == 0)
    {
        if (row == 0)
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME  ) ;
            row = 1 ;
        }
        else
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME ) ;
            row = 0 ;
        }

        //Now shift to the end of the row
        for (i = 0 ; i < ( LCD_MAX_COLUMN - 1 ) ; i++)
        {
            LCD_ShiftCursorRight ( ) ;
        }
    }
    else
    {
        column-- ;
        LCD_SendCommand ( LCD_COMMAND_MOVE_CURSOR_LEFT ) ;
    }
}
/*********************************************************************
 * Function: static void LCD_ShiftCursorRight(void)
 *
 * Overview: Shifts cursor right one spot (wrapping if required)
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
static void LCD_ShiftCursorRight ( void )
{
    LCD_SendCommand ( LCD_COMMAND_MOVE_CURSOR_RIGHT ) ;
    column++ ;

    if (column == LCD_MAX_COLUMN)
    {
        column = 0 ;
        if (row == 0)
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_1_HOME  ) ;
            row = 1 ;
        }
        else
        {
            LCD_SendCommand ( LCD_COMMAND_ROW_0_HOME ) ;
            row = 0 ;
        }
    }
}
/*********************************************************************
 * Function: static void LCD_ShiftCursorUp(void)
 *
 * Overview: Shifts cursor up one spot (wrapping if required)
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
static void LCD_ShiftCursorUp ( void )
{
    uint8_t i ;

    for (i = 0 ; i < LCD_MAX_COLUMN ; i++)
    {
        LCD_ShiftCursorLeft ( ) ;
    }
}
/*********************************************************************
 * Function: static void LCD_ShiftCursorDown(void)
 *
 * Overview: Shifts cursor down one spot (wrapping if required)
 *
 * PreCondition: already initialized via LCD_Initialize()
 *
 * Input: None
 *
 * Output: None
 *
 ********************************************************************/
static void LCD_ShiftCursorDown ( void )
{
    uint8_t i ;

    for (i = 0 ; i < LCD_MAX_COLUMN ; i++)
    {
        LCD_ShiftCursorRight ( ) ;
    }
}
